<div
  style="display: flex; flex-direction: {{{widget._flex_direction}}}; gap: {{{widget._gap}}}px"
>
  {% for item in widget._widgets %} {% if widget._fractions is none %}
  <div>{{{item}}}</div>
  {% else %}
  <div style="flex: {{{widget._fractions[loop.index0]}}}">{{{item}}}</div>
  {% endif %} {% endfor %}
</div>
